import os
import distutils.version
import urllib2
import json

try:
    import urlparse
except ImportError:
    import urllib.parse as urlparse #py3k

#TODO: Delete this module.  This is a temporary experiment

GITHUB_V2_API_REPOS = 'https://github.com/api/v2/json/repos/'
PATTERN_GITHUB_V2_API_REPOS_SHOW_TAGS = GITHUB_V2_API_REPOS + 'show/%(org_name)s/%(repo_name)s/tags'
# include extra filename on the end to help brew with its infererence rules.  Github should ignore them.
PATTERN_GITHUB_TARBALL_DOWNLOAD = 'https://github.com/%(org_name)s/%(repo_name)s/tarball/%(tag_name)s/%(repo_name)s-%(version)s.tar.gz'

def get_api(api_pattern, org_name, repo_name):
    return api_pattern%locals()

def list_tags(org_name, repo_name, prefix):
    f = urllib2.urlopen(get_api(PATTERN_GITHUB_V2_API_REPOS_SHOW_TAGS, org_name, repo_name))
    json_data = json.load(f)
    f.close()
    return [t for t in json_data['tags'].keys() if t.startswith(prefix)]

def get_org_name(url):
    parsed = urlparse.urlparse(url)
    org_name = os.path.dirname(parsed.path)
    if org_name.startswith('git@github.com:'):
        org_name = org_name[len('git@github.com:'):]
    return org_name

def get_repo_name(url):
    repo_name = os.path.basename(url)
    if repo_name.endswith('.git'):
        repo_name = repo_name[:-4]
    return repo_name

def compute_tarball_url_for_latest(org_name, repo_name):
    """
    Compute the latest upstream tag and return a github.com download URL for that tarball
    """
    prefix = 'upstream/'
    tags = list_tags(org_name, repo_name, prefix)
    versions = sorted([distutils.version.LooseVersion(t[len(prefix):]) for t in tags])
    version = versions[-1].vstring #for pattern
    tag_name = 'upstream/%s'%(version)
    return PATTERN_GITHUB_TARBALL_DOWNLOAD%locals()




#!/usr/bin/env python
# Copyright (c) 2009, Willow Garage, Inc.
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the Willow Garage, Inc. nor the names of its
#       contributors may be used to endorse or promote products derived from
#       this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Author Tully Foote, Ken Conley

from rospkg.os_detect import OS_DEBIAN, OS_UBUNTU

from .pip import PIP_INSTALLER
from .gem import GEM_INSTALLER
from .source import SOURCE_INSTALLER
from ..installers import PackageManagerInstaller, TYPE_CODENAME
from ..shell_utils import read_stdout

# apt package manager key
APT_INSTALLER='apt'

def register_installers(context):
    context.set_installer(APT_INSTALLER, AptInstaller())

def register_platforms(context):
    register_debian(context)
    register_ubuntu(context)
    
def register_debian(context):
    context.add_os_installer_key(OS_DEBIAN, APT_INSTALLER)
    context.add_os_installer_key(OS_DEBIAN, PIP_INSTALLER)
    context.add_os_installer_key(OS_DEBIAN, GEM_INSTALLER)
    context.add_os_installer_key(OS_DEBIAN, SOURCE_INSTALLER)
    context.set_default_os_installer_key(OS_DEBIAN, APT_INSTALLER)
    context.set_os_version_type(OS_DEBIAN, TYPE_CODENAME)
    
def register_ubuntu(context):
    context.add_os_installer_key(OS_UBUNTU, APT_INSTALLER)
    context.add_os_installer_key(OS_UBUNTU, PIP_INSTALLER)
    context.add_os_installer_key(OS_UBUNTU, GEM_INSTALLER)
    context.add_os_installer_key(OS_UBUNTU, SOURCE_INSTALLER)
    context.set_default_os_installer_key(OS_UBUNTU, APT_INSTALLER)
    context.set_os_version_type(OS_UBUNTU, TYPE_CODENAME)

def dpkg_detect(pkgs, exec_fn=None):
    """ 
    Given a list of package, return the list of installed packages.

    :param exec_fn: function to execute Popen and read stdout (for testing)
    """
    ret_list = []
    # this is mainly a hack to support version locking for eigen.
    # we strip version-locking syntax, e.g. libeigen3-dev=3.0.1-*.
    # our query does not do the validation on the version itself.
    version_lock_map = {}
    for p in pkgs:
        if '=' in p:
            version_lock_map[p.split('=')[0]] = p
        else:
            version_lock_map[p] = p
    cmd = ['dpkg-query', '-W', '-f=\'${Package} ${Status}\n\'']
    cmd.extend(version_lock_map.keys())

    if exec_fn is None:
        exec_fn = read_stdout
    std_out = exec_fn(cmd)
    std_out = std_out.replace('\'','')
    pkg_list = std_out.split('\n')
    for pkg in pkg_list:
        pkg_row = pkg.split()
        if len(pkg_row) == 4 and (pkg_row[3] =='installed'):
            ret_list.append( pkg_row[0])
    return [version_lock_map[r] for r in ret_list]

class AptInstaller(PackageManagerInstaller):
    """ 
    An implementation of the Installer for use on debian style
    systems.
    """
    def __init__(self):
        super(AptInstaller, self).__init__(dpkg_detect)

    def get_install_command(self, resolved, interactive=True, reinstall=False):
        packages = self.get_packages_to_install(resolved, reinstall=reinstall)
        if not packages:
            return []
        if not interactive:
            return [['sudo', 'apt-get', 'install', '-y', p] for p in packages]
        else:
            return [['sudo', 'apt-get', 'install', p] for p in packages]



# Copyright (c) 2009, Willow Garage, Inc.
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the Willow Garage, Inc. nor the names of its
#       contributors may be used to endorse or promote products derived from
#       this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Author Tully Foote/tfoote@willowgarage.com

import subprocess

from rospkg.os_detect import OS_OPENSUSE

from .source import SOURCE_INSTALLER
from ..installers import PackageManagerInstaller

# zypper package manager key
ZYPPER_INSTALLER='zypper'

def register_installers(context):
    context.set_installer(ZYPPER_INSTALLER, ZypperInstaller())

def register_platforms(context):
    context.add_os_installer_key(OS_OPENSUSE, SOURCE_INSTALLER)
    context.add_os_installer_key(OS_OPENSUSE, ZYPPER_INSTALLER)
    context.set_default_os_installer_key(OS_OPENSUSE, ZYPPER_INSTALLER)
    
def rpm_detect(packages):
    return subprocess.call(['rpm', '-q', packages], stdout=subprocess.PIPE, stderr=subprocess.PIPE)    

class ZypperInstaller(PackageManagerInstaller):
    """
    This class provides the functions for installing using zypper.
    """

    def __init__(self):
        super(ZypperInstaller, self).__init__(rpm_detect)

    def get_install_command(self, resolved, interactive=True, reinstall=False):
        packages = self.get_packages_to_install(resolved, reinstall=reinstall)
        if not packages:
            return []
        if not interactive:
            return [['sudo', 'zypper', 'install', '-yl']+packages]
        else:
            return [['sudo', 'zypper', 'install']+packages]



import urllib2
import yaml
import urlparse

from rospkg.os_detect import OS_UBUNTU
from rospkg.os_detect import OS_OSX

from .core import InvalidData, DownloadFailure
from .platforms.debian import APT_INSTALLER
from .platforms.osx import BREW_INSTALLER
from .rep3 import download_targets_data

#py3k
try:
    unicode
except:
    basestring = unicode = str

# location of an example gbpdistro file for reference and testing
FUERTE_GBPDISTRO_URL = 'https://raw.github.com/ros/rosdistro/' \
                     + 'master/releases/fuerte.yaml'

#seconds to wait before aborting download of gbpdistro data
DOWNLOAD_TIMEOUT = 15.0


def get_owner_name(url):
    """
    Given a gbpdistro url, returns the name of the github user in the url.

    If the url is not a valid github url it returns the default `ros`.

    This information is used to set the homebrew tap name, see:
    https://github.com/ros/rosdep/pull/17

    :returns: The github account in the given gbpdistro url
    """
    result = 'ros'
    try:
        parsed = urlparse.urlparse(url)
        if parsed.netloc == 'github.com':
            result = parsed.path.split('/')[1]
    except Exception:
        pass
    return result


# For compatability url defaults to ''
def gbprepo_to_rosdep_data(gbpdistro_data, targets_data, url=''):
    """
    :raises: :exc:`InvalidData`
    """
    # Error reporting for this isn't nearly as good as it could be
    # (e.g. doesn't separate gbpdistro vs. targets, nor provide
    # origin), but rushing this implementation a bit.
    try:
        if not type(targets_data) == dict:
            raise InvalidData("targets data must be a dict")
        if not type(gbpdistro_data) == dict:
            raise InvalidData("gbpdistro data must be a dictionary")
        if gbpdistro_data['type'] != 'gbp':
            raise InvalidData('gbpdistro must be of type "gbp"')

        # compute the default target data for the release_name
        release_name = gbpdistro_data['release-name']
        if not release_name in targets_data:
            raise InvalidData("targets file does not contain information "
                            + "for release [%s]" % (release_name))
        else:
            # take the first match
            target_data = targets_data[release_name]

        # compute the rosdep data for each repo
        rosdep_data = {}
        gbp_repos = gbpdistro_data['repositories']
        # Ensure gbp_repos is a dict
        if type(gbp_repos) != dict:
            raise InvalidData("invalid repo spec in gbpdistro data: " + str(gbp_repos)
                            + ". Invalid repositories entry, must be dict.")
        for rosdep_key, repo in gbp_repos.items():
            if type(repo) != dict:
                raise InvalidData("invalid repo spec in gbpdistro data: "
                                + str(repo))

            for pkg in repo.get('packages', {rosdep_key: None}):
                rosdep_data[pkg] = {}

                # for pkg in repo['packages']: indent the rest of the lines here.
                # Do generation for ubuntu
                rosdep_data[pkg][OS_UBUNTU] = {}
                # Do generation for empty OS X entries
                homebrew_name = '%s/%s/%s' % (get_owner_name(url),
                                              release_name, rosdep_key)
                rosdep_data[pkg][OS_OSX] = {
                    BREW_INSTALLER: {'packages': [homebrew_name]}
                }

                # - debian package name: underscores must be dashes
                deb_package_name = 'ros-%s-%s' % (release_name, rosdep_key)
                deb_package_name = deb_package_name.replace('_', '-')

                repo_targets = repo['target'] if 'target' in repo else 'all'
                if repo_targets == 'all':
                    repo_targets = target_data

                for t in repo_targets:
                    if not isinstance(t, basestring):
                        raise InvalidData("invalid target spec: %s" % (t))
                    # rosdep_data[pkg][OS_UBUNTU][t] = {
                    rosdep_data[pkg][OS_UBUNTU][t] = {
                        APT_INSTALLER: {'packages': [deb_package_name]}
                    }
        return rosdep_data
    except KeyError as e:
        raise InvalidData("Invalid GBP-distro/targets format: missing key: "
                        + str(e))


def download_gbpdistro_as_rosdep_data(gbpdistro_url, targets_url=None):
    """
    Download gbpdistro file from web and convert format to rosdep distro data.

    :param gbpdistro_url: url of gbpdistro file, ``str``
    :param target_url: override URL of platform targets file
    :raises: :exc:`DownloadFailure`
    :raises: :exc:`InvalidData` If targets file does not pass cursory
     validation checks.
    """
    # we can convert a gbpdistro file into rosdep data by following a
    # couple rules
    targets_data = download_targets_data(targets_url=targets_url)
    try:
        f = urllib2.urlopen(gbpdistro_url, timeout=DOWNLOAD_TIMEOUT)
        text = f.read()
        f.close()
        gbpdistro_data = yaml.safe_load(text)
        return gbprepo_to_rosdep_data(gbpdistro_data,
                                      targets_data,
                                      gbpdistro_url)
    except Exception as e:
        raise DownloadFailure("Failed to download target platform data "
                            + "for gbpdistro:\n\t" + str(e))



# Copyright (c) 2011, Willow Garage, Inc.
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the Willow Garage, Inc. nor the names of its
#       contributors may be used to endorse or promote products derived from
#       this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

def test_RosdepDatabaseEntry():
    # not muich to test with container
    from rosdep2.model import RosdepDatabaseEntry
    d = RosdepDatabaseEntry({'a': 1}, [], 'foo')
    assert d.rosdep_data == {'a': 1}
    assert d.view_dependencies == []
    assert d.origin == 'foo'

def test_RosdepDatabase():
    from rosdep2.model import RosdepDatabase

    db = RosdepDatabase()
    assert not db.is_loaded('foo')

    data = {'a': 1}
    db.set_view_data('foo', data, [], 'origin1')
    assert db.is_loaded('foo')    
    entry = db.get_view_data('foo')
    assert entry.rosdep_data == data
    assert entry.origin == 'origin1'
    assert entry.view_dependencies == []
    # make sure data is copy
    data['a'] = 2
    assert entry.rosdep_data != data
    
    data = {'b': 2}
    db.set_view_data('bar', data, ['foo'], 'origin2')
    assert db.is_loaded('bar')    
    entry = db.get_view_data('bar')
    assert entry.rosdep_data == data
    assert entry.origin == 'origin2'
    assert entry.view_dependencies == ['foo']

    # override entry for bar
    data = {'b': 3}
    assert db.is_loaded('bar')    
    db.set_view_data('bar', data, ['baz', 'blah'], 'origin3')
    assert db.is_loaded('bar')    
    entry = db.get_view_data('bar')
    assert entry.rosdep_data == data
    assert entry.origin == 'origin3'
    assert set(entry.view_dependencies) == set(['baz', 'blah'])
    

def test_RosdepDatabase_get_view_dependencies():
    from rosdep2.model import RosdepDatabase

    data = {'a': 1}
    db = RosdepDatabase()
    
    db.set_view_data('foo', data, [], 'origin')
    assert [] == db.get_view_dependencies('foo')

    db.set_view_data('bar', data, ['foo'], 'origin')
    assert ['foo'] == db.get_view_dependencies('bar')

    db.set_view_data('baz', data, ['bar'], 'origin')
    assert ['foo', 'bar'] ==  db.get_view_dependencies('baz')

    db.set_view_data('rad', data, [], 'origin')
    db.set_view_data('fad', data, ['baz', 'rad'], 'origin')
    retval = db.get_view_dependencies('fad')
    assert set(['baz', 'rad', 'foo', 'bar']) == set(retval), retval
    assert len(retval) == 4



from setuptools import setup

import sys
sys.path.insert(0, 'src')

from rosdep2 import __version__

setup(name='rosdep',
      version= __version__,
      packages=['rosdep2', 'rosdep2.platforms'],
      package_dir = {'':'src'},
#      data_files=[('man/man1', ['doc/man/rosdep.1'])],
      install_requires = ['rospkg', 'PyYAML>=3.1'],
      setup_requires = ['nose>=1.0'],
      test_suite = 'nose.collector',
      test_requires = ['mock'],
      scripts = [
        'scripts/rosdep',
        'scripts/rosdep-gbp-brew',
        'scripts/rosdep-source',
        ],
      author = "Tully Foote, Ken Conley", 
      author_email = "foote@willowgarage.com, kwc@willowgarage.com",
      url = "http://www.ros.org/wiki/rosdep",
      download_url = "http://pr.willowgarage.com/downloads/rosdep/", 
      keywords = ["ROS"],
      classifiers = [
        "Programming Language :: Python", 
        "License :: OSI Approved :: BSD License" ],
      description = "rosdep system dependency installation tool", 
      long_description = """\
Command-line tool for installing system dependencies on a variety of platforms.
""",
      license = "BSD"
      )
